/*
 * thread.cpp
 * Copyright 2005 Linuon Inc. Tokyo Japan All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *     Junji Kanemaru <junji.kanemaru@linuon.com>
 */
#include <unistd.h>
#include <time.h>
#include <errno.h>

#include <string>

#include "fault.h"
#include "thread.h"

using namespace std;

/**
** Static thread entry for pthread_create */
void* Thread::threadEntry(void* _this) {
    pthread_attr_t ta;
    struct sched_param schd_param;

    Thread* th = (Thread*)_this;
    th->stat = TH_RUNNING;
    th->ret = 0;

    pthread_attr_init(&ta);
    pthread_attr_getschedparam(&ta, &schd_param);
    switch (th->prio) {
        case PRIO_HIGH:
            schd_param.sched_priority += 15;
            break;
        case PRIO_LOW:
            schd_param.sched_priority -= 15;
            break;
        case PRIO_NORMAL:
            schd_param.sched_priority += 5;
            break;
    }
    pthread_attr_setschedparam(&ta, &schd_param);

    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
    pthread_cleanup_push(&threadCancelHandler, _this);
    do {
        pthread_testcancel();
        /**
        ** Give pthread library some time to cancel thread 
        ** if there's request */
        usleep(0); 
        if (th->stat == TH_SUSPENDED) {
            pthread_mutex_lock(&th->mutex);
            while (th->stat == TH_SUSPENDED) {
                pthread_cond_wait(&th->cond, &th->mutex);
            }
            pthread_mutex_unlock(&th->mutex);
        }
    } while (th->mainHandler(th->ret, th->userData));

    pthread_cleanup_pop(th->stat == TH_RUNNING);
    th->stat = TH_DONE;
    return &th->ret;
}

/**
** Static thread cancel handler */
void Thread::threadCancelHandler(void* _this) {
    Thread* th = (Thread*)_this;
    th->stat = TH_CANCELED;
}

/**
** Constractors and destractor */
Thread::Thread() { 
    stat = TH_READY; 
    pthread_mutex_init(&mutex, NULL);
    pthread_cond_init(&cond, NULL);
    sleepTime = 0;
}

Thread::~Thread() {
    pthread_mutex_destroy(&mutex);
    pthread_cond_destroy(&cond);
    cancel();
}

void Thread::start(void* userData, Priority prio) {
    this->prio = prio;
    this->userData = userData;

    int rc = pthread_create(&thID, NULL, &threadEntry, (void*)this);
    if (rc != 0) {
        throw SYSTEM_FAULT(rc, strerror(rc));
    }

    pthread_detach(thID);
    while(stat == TH_READY) {
        usleep(5);
    }
}

void Thread::cancel() {
    if (stat == TH_RUNNING) {
        pthread_cancel(thID);
    }

    /* To make sure the thread is all done */
    while(stat == TH_RUNNING) {
        usleep(5);
    }
}

void Thread::suspend() {
    if (stat != TH_RUNNING) {
        return;
    }
    statSaved = stat;
    stat = TH_SUSPENDED;
}

void Thread::resume() {
    if (stat != TH_SUSPENDED) {
        return;
    }
    stat = statSaved;
    pthread_cond_signal(&cond);
    usleep(5);
}

int Thread::wait(time_t second) {
    time_t t;

    t = time(NULL);
    while(stat == TH_RUNNING && 
            (second == 0 ? true : ((time(NULL) - t) < second))) {
        usleep(100);
    }

    /* If timed out */
    if (stat != TH_DONE) {
        cancel();
        while(stat == TH_RUNNING) { 
            usleep(5);
        }
    }

    return ret;
}

